from typing import Any, Dict, List, Optional, Union

import httpx

from ...client import Client
from ...models.training_set_version import TrainingSetVersion
from ...models.training_set_version_filter import TrainingSetVersionFilter
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *,
    client: Client,
    json_body: TrainingSetVersionFilter,
    offset: Union[Unset, None, int] = UNSET,
    limit: Union[Unset, None, int] = UNSET,
    asc: Union[Unset, None, bool] = UNSET,
) -> Dict[str, Any]:
    url = "{}/version/find".format(client.base_url)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    params: Dict[str, Any] = {
        "offset": offset,
        "limit": limit,
        "asc": asc,
    }
    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    json_json_body = json_body.to_dict()

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "json": json_json_body,
        "params": params,
        "verify": client.verify_ssl,
    }


def _parse_response(*, response: httpx.Response) -> Optional[List[TrainingSetVersion]]:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = TrainingSetVersion.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[List[TrainingSetVersion]]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
    json_body: TrainingSetVersionFilter,
    offset: Union[Unset, None, int] = UNSET,
    limit: Union[Unset, None, int] = UNSET,
    asc: Union[Unset, None, bool] = UNSET,
) -> Response[List[TrainingSetVersion]]:
    kwargs = _get_kwargs(
        client=client,
        json_body=json_body,
        offset=offset,
        limit=limit,
        asc=asc,
    )

    response = httpx.post(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
    json_body: TrainingSetVersionFilter,
    offset: Union[Unset, None, int] = UNSET,
    limit: Union[Unset, None, int] = UNSET,
    asc: Union[Unset, None, bool] = UNSET,
) -> Optional[List[TrainingSetVersion]]:
    """ """

    return sync_detailed(
        client=client,
        json_body=json_body,
        offset=offset,
        limit=limit,
        asc=asc,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    json_body: TrainingSetVersionFilter,
    offset: Union[Unset, None, int] = UNSET,
    limit: Union[Unset, None, int] = UNSET,
    asc: Union[Unset, None, bool] = UNSET,
) -> Response[List[TrainingSetVersion]]:
    kwargs = _get_kwargs(
        client=client,
        json_body=json_body,
        offset=offset,
        limit=limit,
        asc=asc,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.post(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
    json_body: TrainingSetVersionFilter,
    offset: Union[Unset, None, int] = UNSET,
    limit: Union[Unset, None, int] = UNSET,
    asc: Union[Unset, None, bool] = UNSET,
) -> Optional[List[TrainingSetVersion]]:
    """ """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
            offset=offset,
            limit=limit,
            asc=asc,
        )
    ).parsed



""" Contains all the data models used in inputs/outputs """

from .datasource_config import DatasourceConfig
from .datasource_dto import DatasourceDto
from .datasource_dto_added_by import DatasourceDtoAddedBy
from .datasource_dto_auth_type import DatasourceDtoAuthType
from .datasource_dto_data_source_type import DatasourceDtoDataSourceType
from .datasource_dto_status import DatasourceDtoStatus
from .datasource_owner_info import DatasourceOwnerInfo
from .error_response import ErrorResponse
from .key_request import KeyRequest
from .list_request import ListRequest
from .log_metric_m import LogMetricM
from .log_metric_t import LogMetricT
from .proxy_error_response import ProxyErrorResponse



"""Containing APIs to sync feature store from feast to domino"""

import os
import time
from pathlib import Path

from git import Repo

from feature_store_api_client.models import (
    Entity,
    Feature,
    FeatureStoreSyncResult,
    FeatureTags,
    FeatureViewRequest,
    FeatureViewRequestTags,
    LockFeatureStoreRequest,
    UnlockFeatureStoreRequest,
)
from feature_store_api_client.types import UNSET

from .client import FeatureStoreClient
from .exceptions import FeastRepoError, FeatureStoreLockError
from .git import pull_repo, push_to_git
from .logging import logger

LOCK_FAILURE_MESSAGE = "Failed to lock feature store for syncing. Please rerun later."

_import_error_message = (
    "feast is not installed.\n\n"
    "Please pip install feast:\n\n"
    "  python -m pip install feast   # install directly\n"
)

try:
    from feast import FeatureStore, repo_config, repo_operations
except ImportError as e:
    if e.msg == "No module named 'feast'":
        raise ImportError(_import_error_message) from e
    else:
        raise


def lock(feature_store_id: str, max_retries: int) -> None:
    """Lock the feature store for updating features. If fails, retry until
    succeeds or reaches the maximum retry limit
    Args:
        feature_store_id: the id of the feature store to be locked
        max_retries: the maximum number of retries
    Raises:
        FeatureStoreLockError: if fails to lock
    """
    logger.info("Locking the feature store")
    client = FeatureStoreClient()
    lock_request = LockFeatureStoreRequest(
        feature_store_id=feature_store_id,
        project_name=os.getenv("DOMINO_PROJECT_NAME", ""),
        user_name=os.getenv("DOMINO_STARTING_USERNAME", ""),
        run_id=os.getenv("DOMINO_RUN_ID", ""),
    )
    lock_result = False
    retry_count = 0
    while not lock_result and retry_count <= max_retries:
        lock_result = client.lock(lock_request)
        if not lock_result:
            time.sleep(1)
            logger.info(f"Failed to lock the feature store. retry count: {retry_count}")
            retry_count += 1

    if not lock_result:
        raise FeatureStoreLockError(LOCK_FAILURE_MESSAGE)
    logger.info("Locked the feature store")


def unlock(feature_store_id: str, sync_result: FeatureStoreSyncResult) -> None:
    """UnLock the feature store
    Args:
        feature_store_id: the id of the feature store to be unlocked
        sync_result: the synchronization result
    """
    logger.info("UnLocking the feature store")
    try:
        client = FeatureStoreClient()
        result = client.unlock(
            UnlockFeatureStoreRequest(feature_store_id=feature_store_id, sync_result=sync_result)
        )
        logger.info(
            "Unlocked the feature store." if result else "Failed to unlock the feature store."
        )
    except Exception as e:
        logger.exception(e)


def update_feature_views(commit_id: str, repo_path: str) -> None:
    """Update domino feature views to sync with specified feast git commit
    Args:
        commit_id: the feast git repo commit id that are to be synced with domino
        repo_path: the feast git repo path
    """
    logger.info(f"Syncing feature views to feast git commit {commit_id}")

    feast_feature_store = FeatureStore(repo_path)

    feature_views = feast_feature_store.list_feature_views()

    request_input = []
    for fv in feature_views:
        feature_v = FeatureViewRequest(
            name=fv.name,
            entities=[
                Entity(name=x, join_key=y.name, value_type=str(y.dtype))
                for x, y in zip(fv.entities, fv.entity_columns)
            ],
            features=[
                Feature(name=f.name, dtype=str(f.dtype), tags=FeatureTags.from_dict(f.tags))
                for f in fv.features
            ],
            ttl=UNSET if fv.ttl is None else int(fv.ttl.total_seconds() * 1000),
            tags=FeatureViewRequestTags.from_dict(fv.tags),
            description=fv.description,
        )
        request_input.append(feature_v)

    client = FeatureStoreClient()
    client.post_feature_views(
        request_input,
        commit_id,
        os.getenv("DOMINO_PROJECT_ID", ""),
    )
    logger.info("Feature Views successfully synced.")


def find_feast_repo_path(root_dir: str) -> str:
    """Find the feast repo path
    Args:
        root_dir: the feast git repo root directory
    Returns:
        the feast repo path
    Raises:
        FeastRepoError: if no feast repo or more than one repo in the specified root directory
        FileNotFoundError: the root path doesn't exist
        NotADirectoryError: the root path is not a directory
    """
    sub_dirs = []
    if not os.path.exists(root_dir):
        raise FileNotFoundError(f"The repo root path {root_dir} does not exist.")
    if not os.path.isdir(root_dir):
        raise NotADirectoryError(f"The repo root path {root_dir} is not a directory.")

    for file_name in os.listdir(root_dir):
        if file_name.startswith("."):
            continue
        full_path = os.path.join(root_dir, file_name)
        if os.path.isdir(full_path):
            sub_dirs.append(full_path)

    if len(sub_dirs) == 0:
        raise FeastRepoError(f"No repo is found under {root_dir}")

    if len(sub_dirs) > 1:
        raise FeastRepoError(f"Multiple repos found under {root_dir}: " + ",".join(sub_dirs))

    return sub_dirs[0]


def run_feast_apply(repo_path_str: str, skip_source_validation: bool = False) -> None:
    """run feast apply
    Args:
        repo_path_str: the feast git repo path
        skip_source_validation: don't validate the data sources by
                                checking for that the tables exist if true
    """
    logger.info("running feast apply")
    repo_path = Path(repo_path_str).absolute()
    fs_yaml_file = repo_path / "feature_store.yaml"
    repo_operations.cli_check_repo(repo_path, fs_yaml_file)

    feast_repo_config = repo_config.load_repo_config(repo_path, fs_yaml_file)
    repo_operations.apply_total(feast_repo_config, repo_path, skip_source_validation)


def feature_store_sync(
    feature_store_id: str,
    repo_path_str: str,
    branch_name: str,
    max_retries: int,
    skip_source_validation: bool = False,
) -> None:
    """run feature store syncing
    Args:
        feature_store_id: the feature store domino id
        repo_path_str: feast repo path
        branch_name: feast repo branch
        max_retries: maximum lock retry count
        skip_source_validation: option for running feast apply command.
                                Don't validate the data sources by checking
                                for that the tables exist if true
    """
    if not repo_path_str:
        root_dir = os.getenv("DOMINO_FEAST_REPO_ROOT", "/features")
        repo_path_str = find_feast_repo_path(root_dir)

    repo = Repo(repo_path_str)
    if not branch_name:
        branch_name = repo.active_branch.name

    if not max_retries:
        max_retries = 60

    logger.info(
        f"Starting syncing for feature store {feature_store_id} with {repo_path_str} "
        f"on branch {branch_name}"
    )

    result = FeatureStoreSyncResult.FAILURE
    lock(feature_store_id, max_retries)
    try:
        # Feature store syncing is going to be run in a domino job. Turn off it
        # to avoid the user interaction on host authentication. Feast git repo
        # has already been pulled and mounted as a volume to the job pod, so the host
        # can be trusted.
        git_ssh_cmd = "ssh -o StrictHostKeyChecking=no"
        with repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
            pull_repo(repo, branch_name)
            run_feast_apply(
                repo_path_str=repo_path_str, skip_source_validation=skip_source_validation
            )
            push_to_git(repo)

        update_feature_views(repo.head.object.hexsha, repo_path_str)
        result = FeatureStoreSyncResult.SUCCESS
        logger.info("Finished feature store syncing.")
    finally:
        unlock(feature_store_id, result)



from enum import Enum


class GitProviderName(str, Enum):
    BITBUCKET = "Bitbucket"
    BITBUCKETSERVER = "BitbucketServer"
    GITHUB = "Github"
    GITHUBENTERPRISE = "GithubEnterprise"
    GITLAB = "Gitlab"
    GITLABENTERPRISE = "GitlabEnterprise"
    UNKNOWN = "Unknown"

    def __str__(self) -> str:
        return str(self.value)



"""Containing APIs to interact with feast git repository"""

from git import FetchInfo, PushInfo, Repo

from .exceptions import GitPullError, GitPushError
from .logging import logger

_INVALID_PULL_FLAGS = [FetchInfo.ERROR, FetchInfo.REJECTED]
_INVALID_PUSH_FLAGS = [
    PushInfo.ERROR,
    PushInfo.REJECTED,
    PushInfo.REMOTE_FAILURE,
    PushInfo.REMOTE_REJECTED,
]


def pull_repo(repo: Repo, branch_name: str) -> None:
    """Pull feast repo.

    Args:
        repo: the feast Git Repo object.
        branch_name: the branch to pull

    Raises:
        GitPullError: if fails to pull the repo
    """
    current_branch = repo.active_branch.name
    if current_branch == branch_name:
        logger.info(f"The git repo is already on branch {branch_name}")
        logger.info(f"Commit Sha before pulling:{repo.head.object.hexsha}")
        pull_result = repo.remotes.origin.pull()
        result_flag = pull_result[0].flags
        if result_flag in _INVALID_PULL_FLAGS:
            raise GitPullError(f"Failed to pull the repo with error flag {result_flag}")
        logger.info("Finished pulling the repo.")
        logger.info(f"Commit Sha after pulling:{repo.head.object.hexsha}")
    else:
        logger.info(f"Switching branch from {current_branch} to {branch_name}")
        result = repo.git.checkout(branch_name)
        logger.info(result)
        logger.info(f"Switched to branch {branch_name}, commit Sha:{repo.head.object.hexsha}")


def push_to_git(repo: Repo) -> None:
    """Push feast apply result to the feast repo.

    Args:
        repo: The feast Git Repo object.

    Raises:
        GitPushError: if fails to push to the repo
    """
    logger.info("Starting Git add/commit/push......")
    current_commit_id = repo.head.object.hexsha
    registry_file = "data/registry.db"
    registry_file_added = False

    # New added files. e.g. online.db is created when running feast apply for the first time
    for new_added_file in repo.untracked_files:
        repo.git.add(new_added_file)
        logger.info(f"Added new file {new_added_file} to git")

    for item in repo.index.diff(None):
        repo.git.add(item.a_path)
        logger.info(f"Added modified file {item.a_path} to git")
        if item.a_path == registry_file:
            registry_file_added = True

    # This file can be added in .gitignore file. So forcing adding it here.
    if not registry_file_added:
        repo.git.add(registry_file, force=True)
        logger.info(f"Forced adding {registry_file} to git")

    repo.git.commit("-m", f"Domino run feast apply on {current_commit_id}")
    logger.info(f"Committed feast apply result on {current_commit_id}")
    push_result = repo.remotes.origin.push()
    result_flag = push_result[0].flags
    if result_flag in _INVALID_PUSH_FLAGS:
        raise GitPushError(f"Failed to push to the repo with error flag {result_flag}")
    logger.info(f"Pushed to the repo. The commit id {repo.head.object.hexsha}")



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="LockFeatureStoreRequest")


@attr.s(auto_attribs=True)
class LockFeatureStoreRequest:
    """
    Attributes:
        feature_store_id (str):
        project_name (Union[Unset, str]):
        user_name (Union[Unset, str]):
        run_id (Union[Unset, str]):
    """

    feature_store_id: str
    project_name: Union[Unset, str] = UNSET
    user_name: Union[Unset, str] = UNSET
    run_id: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        feature_store_id = self.feature_store_id
        project_name = self.project_name
        user_name = self.user_name
        run_id = self.run_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "featureStoreId": feature_store_id,
            }
        )
        if project_name is not UNSET:
            field_dict["projectName"] = project_name
        if user_name is not UNSET:
            field_dict["userName"] = user_name
        if run_id is not UNSET:
            field_dict["runId"] = run_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        feature_store_id = d.pop("featureStoreId")

        project_name = d.pop("projectName", UNSET)

        user_name = d.pop("userName", UNSET)

        run_id = d.pop("runId", UNSET)

        lock_feature_store_request = cls(
            feature_store_id=feature_store_id,
            project_name=project_name,
            user_name=user_name,
            run_id=run_id,
        )

        lock_feature_store_request.additional_properties = d
        return lock_feature_store_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import datetime

import attr
from dateutil.parser import isoparse

from ..models.monitoring_meta import MonitoringMeta
from ..models.training_set_version_meta import TrainingSetVersionMeta
from ..types import UNSET, Unset

T = TypeVar("T", bound="TrainingSetVersion")


@attr.s(auto_attribs=True)
class TrainingSetVersion:
    """ """

    id: str
    training_set_id: str
    training_set_name: str
    number: int
    creation_time: datetime.datetime
    path: str
    container_path: str
    key_columns: List[str]
    target_columns: List[str]
    exclude_columns: List[str]
    all_columns: List[str]
    monitoring_meta: MonitoringMeta
    meta: TrainingSetVersionMeta
    pending: bool
    description: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        training_set_id = self.training_set_id
        training_set_name = self.training_set_name
        number = self.number
        creation_time = self.creation_time.isoformat()

        path = self.path
        container_path = self.container_path
        key_columns = self.key_columns

        target_columns = self.target_columns

        exclude_columns = self.exclude_columns

        all_columns = self.all_columns

        monitoring_meta = self.monitoring_meta.to_dict()

        meta = self.meta.to_dict()

        pending = self.pending
        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "trainingSetId": training_set_id,
                "trainingSetName": training_set_name,
                "number": number,
                "creationTime": creation_time,
                "path": path,
                "containerPath": container_path,
                "keyColumns": key_columns,
                "targetColumns": target_columns,
                "excludeColumns": exclude_columns,
                "allColumns": all_columns,
                "monitoringMeta": monitoring_meta,
                "meta": meta,
                "pending": pending,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        training_set_id = d.pop("trainingSetId")

        training_set_name = d.pop("trainingSetName")

        number = d.pop("number")

        creation_time = isoparse(d.pop("creationTime"))

        path = d.pop("path")

        container_path = d.pop("containerPath")

        key_columns = cast(List[str], d.pop("keyColumns"))

        target_columns = cast(List[str], d.pop("targetColumns"))

        exclude_columns = cast(List[str], d.pop("excludeColumns"))

        all_columns = cast(List[str], d.pop("allColumns"))

        monitoring_meta = MonitoringMeta.from_dict(d.pop("monitoringMeta"))

        meta = TrainingSetVersionMeta.from_dict(d.pop("meta"))

        pending = d.pop("pending")

        description = d.pop("description", UNSET)

        training_set_version = cls(
            id=id,
            training_set_id=training_set_id,
            training_set_name=training_set_name,
            number=number,
            creation_time=creation_time,
            path=path,
            container_path=container_path,
            key_columns=key_columns,
            target_columns=target_columns,
            exclude_columns=exclude_columns,
            all_columns=all_columns,
            monitoring_meta=monitoring_meta,
            meta=meta,
            pending=pending,
            description=description,
        )

        training_set_version.additional_properties = d
        return training_set_version

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict

import httpx

from ...client import Client
from ...models.log_metric_m import LogMetricM
from ...models.log_metric_t import LogMetricT
from ...types import UNSET, Response


def _get_kwargs(
    *,
    client: Client,
    t: LogMetricT,
    b: int,
    m: LogMetricM,
) -> Dict[str, Any]:
    url = "{}/objectstore/metric".format(client.base_url)

    headers: Dict[str, str] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    params: Dict[str, Any] = {}
    json_t = t.value

    params["t"] = json_t

    params["b"] = b

    json_m = m.value

    params["m"] = json_m

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "method": "head",
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "params": params,
    }


def _build_response(*, response: httpx.Response) -> Response[Any]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=None,
    )


def sync_detailed(
    *,
    client: Client,
    t: LogMetricT,
    b: int,
    m: LogMetricM,
) -> Response[Any]:
    """Log metrics about file size read or written

    Args:
        t (LogMetricT):
        b (int):
        m (LogMetricM):

    Returns:
        Response[Any]
    """

    kwargs = _get_kwargs(
        client=client,
        t=t,
        b=b,
        m=m,
    )

    response = httpx.request(
        verify=client.verify_ssl,
        **kwargs,
    )

    return _build_response(response=response)


async def asyncio_detailed(
    *,
    client: Client,
    t: LogMetricT,
    b: int,
    m: LogMetricM,
) -> Response[Any]:
    """Log metrics about file size read or written

    Args:
        t (LogMetricT):
        b (int):
        m (LogMetricM):

    Returns:
        Response[Any]
    """

    kwargs = _get_kwargs(
        client=client,
        t=t,
        b=b,
        m=m,
    )

    async with httpx.AsyncClient(verify=client.verify_ssl) as _client:
        response = await _client.request(**kwargs)

    return _build_response(response=response)



""" Contains all the data models used in inputs/outputs """

from .auth_config import AuthConfig
from .auth_config_fields import AuthConfigFields
from .auth_config_meta import AuthConfigMeta
from .auth_field_name import AuthFieldName
from .auth_type import AuthType
from .config_field_name import ConfigFieldName
from .create_feature_store_request import CreateFeatureStoreRequest
from .create_feature_store_request_offline_store_config import (
    CreateFeatureStoreRequestOfflineStoreConfig,
)
from .entity import Entity
from .feature import Feature
from .feature_store import FeatureStore
from .feature_store_offline_store_config import FeatureStoreOfflineStoreConfig
from .feature_store_sync_result import FeatureStoreSyncResult
from .feature_tags import FeatureTags
from .feature_view import FeatureView
from .feature_view_request import FeatureViewRequest
from .feature_view_request_tags import FeatureViewRequestTags
from .feature_view_tags import FeatureViewTags
from .field import Field
from .git_provider_name import GitProviderName
from .lock_feature_store_request import LockFeatureStoreRequest
from .metadata import Metadata
from .offline_store_config import OfflineStoreConfig
from .offline_store_config_fields import OfflineStoreConfigFields
from .offline_store_type import OfflineStoreType
from .unlock_feature_store_request import UnlockFeatureStoreRequest
from .upsert_feature_views_request import UpsertFeatureViewsRequest



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.datasource_config import DatasourceConfig
from ..types import UNSET, Unset

T = TypeVar("T", bound="ListRequest")


@attr.s(auto_attribs=True)
class ListRequest:
    """
    Attributes:
        datasource_id (str):
        prefix (str):
        config_overwrites (Union[Unset, DatasourceConfig]):
        credential_overwrites (Union[Unset, DatasourceConfig]):
        page_size (Union[Unset, int]):
    """

    datasource_id: str
    prefix: str
    config_overwrites: Union[Unset, DatasourceConfig] = UNSET
    credential_overwrites: Union[Unset, DatasourceConfig] = UNSET
    page_size: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        datasource_id = self.datasource_id
        prefix = self.prefix
        config_overwrites: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.config_overwrites, Unset):
            config_overwrites = self.config_overwrites.to_dict()

        credential_overwrites: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.credential_overwrites, Unset):
            credential_overwrites = self.credential_overwrites.to_dict()

        page_size = self.page_size

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "datasourceId": datasource_id,
                "prefix": prefix,
            }
        )
        if config_overwrites is not UNSET:
            field_dict["configOverwrites"] = config_overwrites
        if credential_overwrites is not UNSET:
            field_dict["credentialOverwrites"] = credential_overwrites
        if page_size is not UNSET:
            field_dict["page_size"] = page_size

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        datasource_id = d.pop("datasourceId")

        prefix = d.pop("prefix")

        _config_overwrites = d.pop("configOverwrites", UNSET)
        config_overwrites: Union[Unset, DatasourceConfig]
        if isinstance(_config_overwrites, Unset):
            config_overwrites = UNSET
        else:
            config_overwrites = DatasourceConfig.from_dict(_config_overwrites)

        _credential_overwrites = d.pop("credentialOverwrites", UNSET)
        credential_overwrites: Union[Unset, DatasourceConfig]
        if isinstance(_credential_overwrites, Unset):
            credential_overwrites = UNSET
        else:
            credential_overwrites = DatasourceConfig.from_dict(_credential_overwrites)

        page_size = d.pop("page_size", UNSET)

        list_request = cls(
            datasource_id=datasource_id,
            prefix=prefix,
            config_overwrites=config_overwrites,
            credential_overwrites=credential_overwrites,
            page_size=page_size,
        )

        list_request.additional_properties = d
        return list_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="Entity")


@attr.s(auto_attribs=True)
class Entity:
    """
    Attributes:
        name (str):
        value_type (str):
        join_key (Union[Unset, str]):
    """

    name: str
    value_type: str
    join_key: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        value_type = self.value_type
        join_key = self.join_key

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "valueType": value_type,
            }
        )
        if join_key is not UNSET:
            field_dict["joinKey"] = join_key

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        value_type = d.pop("valueType")

        join_key = d.pop("joinKey", UNSET)

        entity = cls(
            name=name,
            value_type=value_type,
            join_key=join_key,
        )

        entity.additional_properties = d
        return entity

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Datasource module."""
from typing import Any, Dict, List, Optional, cast

import configparser
import json
import os

import attr
import backoff
import httpx
import pandas
import urllib3
from pyarrow import ArrowException, flight, parquet

import domino_data.configuration_gen
from datasource_api_client.api.datasource import get_datasource_by_name
from datasource_api_client.api.proxy import get_key_url, list_keys, log_metric
from datasource_api_client.models import DatasourceConfig as APIConfig
from datasource_api_client.models import (
    DatasourceDto,
    DatasourceDtoAuthType,
    DatasourceDtoDataSourceType,
    ErrorResponse,
    KeyRequest,
    ListRequest,
    LogMetricM,
    LogMetricT,
    ProxyErrorResponse,
)

from .auth import AuthenticatedClient, AuthMiddlewareFactory, ProxyClient, get_jwt_token
from .configuration_gen import Config, CredElem, DatasourceConfig, find_datasource_klass
from .logging import logger
from .meta import MetaMiddlewareFactory
from .transfer import MAX_WORKERS, BlobTransfer

ACCEPT_HEADERS = {"Accept": "application/json"}
ADLS_HEADERS = {"X-Ms-Blob-Type": "BlockBlob"}

CREDENTIAL_TYPE = "credential"
CONFIGURATION_TYPE = "configuration"

FLIGHT_ERROR_SPLIT = ". gRPC client debug context:"

AWS_CREDENTIALS_DEFAULT_LOCATION = "/var/lib/domino/home/.aws/credentials"
AWS_SHARED_CREDENTIALS_FILE = "AWS_SHARED_CREDENTIALS_FILE"
DOMINO_API_HOST = "DOMINO_API_HOST"
DOMINO_API_PROXY = "DOMINO_API_PROXY"
DOMINO_CLIENT_SOURCE = "DOMINO_CLIENT_SOURCE"
DOMINO_DATASOURCE_PROXY_HOST = "DOMINO_DATASOURCE_PROXY_HOST"
DOMINO_DATASOURCE_PROXY_FLIGHT_HOST = "DOMINO_DATASOURCE_PROXY_FLIGHT_HOST"
DOMINO_RUN_ID = "DOMINO_RUN_ID"
DOMINO_USER_API_KEY = "DOMINO_USER_API_KEY"
DOMINO_USER_HOST = "DOMINO_USER_HOST"
DOMINO_TOKEN_DEFAULT_LOCATION = "/var/lib/domino/home/.api/token"
DOMINO_TOKEN_FILE = "DOMINO_TOKEN_FILE"


def __getattr__(name: str) -> Any:
    if name.endswith("Config"):
        return getattr(domino_data.configuration_gen, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


def __dir__() -> List[str]:
    confs = filter(lambda x: x.endswith("Config"), dir(domino_data.configuration_gen))
    return list(globals().keys()) + list(confs)


class DominoError(Exception):
    """Base exception for known errors."""


class UnauthenticatedError(DominoError):
    """To handle exponential backoff."""


def _unpack_flight_error(error: str) -> str:
    """Unpack a flight error message by removing extra information."""
    try:
        return error.split(FLIGHT_ERROR_SPLIT, maxsplit=1)[0]
    except ValueError:
        return error


@attr.s
class Result:
    """Represents a query result."""

    client: "DataSourceClient" = attr.ib(repr=False)
    reader: flight.FlightStreamReader = attr.ib(repr=False)
    statement: str = attr.ib()

    def to_pandas(self) -> pandas.DataFrame:
        """Load and transform the result into a pandas DataFrame.

        Returns:
            Pandas dataframe loaded with entire resultset
        """
        return self.reader.read_pandas()

    def to_parquet(self, where: Any) -> None:
        """Load and serialize the result to a local parquet file.

        Args:
            where: path of file-like object.
        """
        table = self.reader.read_all()
        # In Pyarrow v13.0, the parquet version was upgraded to v2.6 from v2.4.
        # Set the coerce_timestamps to "us"(microseconds) for backward compatibility.
        parquet.write_table(table, where, coerce_timestamps="us")


@attr.s
class _Object:
    """Represents an object in a object store."""

    datasource: "ObjectStoreDatasource" = attr.ib(repr=False)
    key: str = attr.ib()

    def http(self) -> httpx.Client:
        """Get datasource http client."""
        return self.datasource.http()

    def pool_manager(self) -> urllib3.PoolManager:
        """Get datasource http pool manager."""
        return self.datasource.pool_manager()

    def get(self) -> bytes:
        """Get object content as bytes."""
        url = self.datasource.get_key_url(self.key, False)
        res = self.http().get(url)
        res.raise_for_status()

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            len(res.content),
            False,
        )

        return res.content

    def download_file(self, filename: str) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exists.

        Args:
            filename: path of file to write content to.
        """
        url = self.datasource.get_key_url(self.key, False)
        content_size = 0
        with self.http().stream("GET", url) as stream, open(filename, "wb") as file:
            for data in stream.iter_bytes():
                content_size += len(data)
                file.write(data)

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            content_size,
            False,
        )

    def download(self, filename: str, max_workers: int = MAX_WORKERS) -> None:
        """Download object content to file with multithreaded support.

        The file will be created if it does not exists. File will be overwritten if it exists.

        Args:
            filename: path of file to write content to
            max_workers: max parallelism for high speed download
        """
        url = self.datasource.get_key_url(self.key, False)
        with open(filename, "wb") as file:
            blob = BlobTransfer(url, file, max_workers=max_workers, http=self.pool_manager())

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            blob.content_size,
            False,
        )

    def download_fileobj(self, fileobj: Any) -> None:
        """Download object content to file like object.

        Args:
            fileobj: A file-like object to download into.
                At a minimum, it must implement the write method and must accept bytes.
        """
        url = self.datasource.get_key_url(self.key, False)
        content_size = 0
        with self.http().stream("GET", url) as stream:
            for data in stream.iter_bytes():
                content_size += len(data)
                fileobj.write(data)

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            content_size,
            False,
        )

    def put(self, content: bytes) -> None:
        """Upload content to object.

        Args:
            content: bytes content
        """
        url = self.datasource.get_key_url(self.key, True)
        res = self.http().put(url, content=content)
        res.raise_for_status()

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            len(content),
            True,
        )

    def upload_file(self, filename: str) -> None:
        """Upload content of file at filename to object.

        Args:
            filename: path of file to upload.
        """
        url = self.datasource.get_key_url(self.key, True)
        with open(filename, "rb") as file:
            res = self.http().put(url, content=file)
        res.raise_for_status()

        content_size = os.path.getsize(filename)
        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            content_size,
            True,
        )

    def upload_fileobj(self, fileobj: Any) -> None:
        """Upload content of file like object to object.

        Args:
            fileobj: bytes-like object or an iterable producing bytes.
        """
        url = self.datasource.get_key_url(self.key, True)
        res = self.http().put(url, content=fileobj)
        res.raise_for_status()


def load_oauth_credentials() -> Dict[str, str]:
    """Load oauth token from sidecar container or local file.

    Returns:
        .. code-block:: python

            {CredElem.TOKEN.value: "token"}

    Raises:
        DominoError: if the provided location is not a valid file
    """
    token_url = os.getenv(DOMINO_API_PROXY, "")
    if token_url:
        try:
            jwt = get_jwt_token(token_url)
        except httpx.HTTPStatusError:
            logger.opt(exception=True).warning("Failed to get token from sidecar container API")
        else:
            return {CredElem.TOKEN.value: jwt}

    location = os.getenv("DOMINO_TOKEN_FILE", DOMINO_TOKEN_DEFAULT_LOCATION)
    try:
        with open(location, encoding="ascii") as token_file:
            return {CredElem.TOKEN.value: token_file.readline().rstrip()}
    except FileNotFoundError as exc:
        raise DominoError("Token file location is not a valid file.") from exc


def load_aws_credentials(location: str, profile: str = "") -> Dict[str, str]:
    """Load AWS credentials from given location and profile.

    Args:
        location: location of file that contains token.
        profile: profile to load.

    Returns:
        .. code-block:: python

            {
                CredElem.ACCESSKEYID.value: "access_key_id",
                CredElem.SECRETACCESSKEY.value: "secret_access_key",
                CredElem.SESSIONTOKEN.value: "session_token",
            }

    Raises:
        DominoError: if the provided location is not a valid file
    """
    location = location or AWS_CREDENTIALS_DEFAULT_LOCATION
    aws_config = configparser.RawConfigParser()
    aws_config.read(location)
    if not aws_config or not aws_config.sections():
        raise DominoError("AWS credentials file does not exist or does not contain profiles")

    profile = profile or aws_config.sections()[0]
    return {
        CredElem.ACCESSKEYID.value: aws_config.get(profile, "aws_access_key_id"),
        CredElem.SECRETACCESSKEY.value: aws_config.get(profile, "aws_secret_access_key"),
        CredElem.SESSIONTOKEN.value: aws_config.get(profile, "aws_session_token"),
    }


@attr.s
class Datasource:
    """Represents a Domino datasource."""

    # pylint: disable=too-many-instance-attributes

    auth_type: str = attr.ib()
    client: "DataSourceClient" = attr.ib(repr=False)
    config: Dict[str, Any] = attr.ib()
    datasource_type: str = attr.ib()
    identifier: str = attr.ib()
    name: str = attr.ib()
    owner: str = attr.ib()

    _config_override: DatasourceConfig = attr.ib(factory=Config, init=False, repr=False)
    _httpx: Optional[httpx.Client] = attr.ib(None, init=False, repr=False)

    @classmethod
    def from_dto(cls, client: "DataSourceClient", dto: DatasourceDto) -> "Datasource":
        """Build a datasource from a given DTO."""
        return cls(
            auth_type=dto.auth_type.value,
            client=client,
            config=dto.config.to_dict(),
            datasource_type=dto.data_source_type.value,
            identifier=dto.id,
            name=dto.name,
            owner=dto.owner_info.owner_name,
        )

    def http(self) -> httpx.Client:
        """Singleton http client built for the datasource."""
        if self._httpx is not None:
            return self._httpx

        context = httpx.create_ssl_context()

        if self.datasource_type in (
            DatasourceDtoDataSourceType.ADLSCONFIG.value,
            DatasourceDtoDataSourceType.AZUREBLOBSTORAGECONFIG.value,
        ):
            self._httpx = httpx.Client(headers=ADLS_HEADERS, verify=context)
        elif self.datasource_type == DatasourceDtoDataSourceType.GENERICS3CONFIG.value:
            self._httpx = httpx.Client(verify=False)  # nosec
        else:
            self._httpx = httpx.Client(verify=context)
        return self._httpx

    def pool_manager(self) -> urllib3.PoolManager:
        """Urllib3 pool manager for range downloads."""
        if self.datasource_type in (
            DatasourceDtoDataSourceType.ADLSCONFIG.value,
            DatasourceDtoDataSourceType.AZUREBLOBSTORAGECONFIG.value,
        ):
            return urllib3.PoolManager(headers=ADLS_HEADERS)
        elif self.datasource_type == DatasourceDtoDataSourceType.GENERICS3CONFIG.value:
            return urllib3.PoolManager(assert_hostname=False)
        else:
            return urllib3.PoolManager()

    def _get_credential_override(self) -> Dict[str, str]:
        """Gets credentials override by merging service overrides and user overrides"""
        credentials = {}

        if self.auth_type == DatasourceDtoAuthType.OAUTH.value:
            credentials = load_oauth_credentials()

        if self.auth_type in (
            DatasourceDtoAuthType.AWSIAMROLE.value,
            DatasourceDtoAuthType.AWSIAMROLEWITHUSERNAME.value,
        ):
            credentials = load_aws_credentials(
                os.getenv(AWS_SHARED_CREDENTIALS_FILE, ""),
                getattr(self._config_override, "profile", ""),
            )

        credentials.update(self._config_override.credential())
        return credentials

    def update(self, config: DatasourceConfig) -> None:
        """Store configuration override for future query calls.

        Args:
            config: specific datasource config class
        """
        self._config_override = config

    def reset_config(self) -> None:
        """Reset the configuration override."""
        self._config_override = Config()


@attr.s
class TabularDatasource(Datasource):
    """Represents a tabular type datasource."""

    def query(self, query: str) -> Result:
        """Execute a query against the datasource.

        Args:
            query: SQL statement to execute

        Returns:
            Result entity wrapping dataframe
        """
        return self.client.execute(
            self.identifier,
            query,
            config=self._config_override.config(),
            credential=self._get_credential_override(),
        )


@attr.s
class ObjectStoreDatasource(Datasource):
    """Represents a object store type datasource."""

    def Object(self, key: str) -> _Object:  # pylint: disable=invalid-name
        """Return an object with given key and datasource client."""
        return _Object(datasource=self, key=key)

    def list_objects(self, prefix: str = "", page_size: int = 1000) -> List[_Object]:
        """List objects in the object store datasource.

        Args:
            prefix: optional prefix to filter objects
            page_size: optional number of objects to fetch

        Returns:
            List of objects
        """
        keys = self.client.list_keys(
            self.identifier,
            prefix,
            page_size,
            config=self._config_override.config(),
            credential=self._get_credential_override(),
        )
        return [
            _Object(
                datasource=self,
                key=key,
            )
            for key in keys
            if not key.endswith("/")
        ]

    def get_key_url(self, object_key: str, is_read_write: bool = False) -> str:
        """Get a signed URL for the given key.

        Args:
            object_key: unique identifier of object to get signed URL for.
            is_read_write: whether the URL should allow writes or not.

        Returns:
            Signed URL for given key
        """
        return cast(
            str,
            self.client.get_key_url(
                self.identifier,
                object_key,
                is_read_write,
                config=self._config_override.config(),
                credential=self._get_credential_override(),
            ),
        )

    def get(self, object_key: str) -> bytes:
        """Get object content as bytes.

        Args:
            object_key: unique key of object

        Returns:
            object content as bytes
        """
        return self.Object(object_key).get()

    def download_file(self, object_key: str, filename: str) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exists.

        Args:
            object_key: unique key of object
            filename: path of file to write content to.
        """
        self.Object(object_key).download_file(filename)

    def download(self, object_key: str, filename: str, max_workers: int = MAX_WORKERS) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exists.

        Args:
            object_key: unique key of object
            filename: path of file to write content to
            max_workers: max parallelism for high speed download
        """
        self.Object(object_key).download(filename, max_workers)

    def download_fileobj(self, object_key: str, fileobj: Any) -> None:
        """Download object content to file like object.

        Args:
            object_key: unique key of object
            fileobj: A file-like object to download into.
                At a minimum, it must implement the write method and must accept bytes.
        """
        self.Object(object_key).download_fileobj(fileobj)

    def put(self, object_key: str, content: bytes) -> None:
        """Upload content to object at given key.

        Args:
            object_key: unique key of object
            content: bytes content
        """
        self.Object(object_key).put(content)

    def upload_file(self, object_key: str, filename: str) -> None:
        """Upload content of file at filename to object at given key.

        Args:
            object_key: unique key of object
            filename: path of file to upload.
        """
        self.Object(object_key).upload_file(filename)

    def upload_fileobj(self, object_key: str, fileobj: Any) -> None:
        """Upload content of file like object to object at given key.

        Args:
            object_key: unique key of object
            fileobj: A file-like object to upload from.
                At a minimum, it must implement the read method and must return bytes.
        """
        self.Object(object_key).upload_fileobj(fileobj)


@attr.s
class BoardingPass:
    """Represent a query request to the Datasource Proxy service."""

    # pylint: disable=too-few-public-methods

    datasource_id: str = attr.ib()
    query: str = attr.ib()
    config: Dict[str, str] = attr.ib()
    credential: Dict[str, str] = attr.ib()

    def to_json(self) -> str:
        """Serialize self to JSON."""
        return json.dumps(
            {
                "datasourceId": self.datasource_id,
                "sqlQuery": self.query,
                "configOverwrites": self.config,
                "credentialOverwrites": self.credential,
            }
        )


@attr.s
class DataSourceClient:
    """API client and bindings."""

    domino: AuthenticatedClient = attr.ib(init=False, repr=False)
    proxy: flight.FlightClient = attr.ib(init=False, repr=False)
    proxy_http: AuthenticatedClient = attr.ib(init=False, repr=False)

    api_key: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_USER_API_KEY))
    token_file: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_TOKEN_FILE))
    token_url: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_API_PROXY))

    def __attrs_post_init__(self):
        flight_host = os.getenv(DOMINO_DATASOURCE_PROXY_FLIGHT_HOST)
        domino_host = os.getenv(
            DOMINO_API_PROXY, os.getenv(DOMINO_API_HOST, os.getenv(DOMINO_USER_HOST, ""))
        )
        proxy_host = os.getenv(DOMINO_DATASOURCE_PROXY_HOST, "")

        client_source = os.getenv(DOMINO_CLIENT_SOURCE, "Python")
        run_id = os.getenv(DOMINO_RUN_ID, "")

        logger.info(
            "initializing datasource client with hosts",
            flight_host=flight_host,
            domino_host=domino_host,
            proxy_host=proxy_host,
        )

        self._set_proxy()
        self.proxy_http = ProxyClient(
            base_url=proxy_host,
            api_key=self.api_key,
            client_source=client_source,
            run_id=run_id,
            token_file=self.token_file,
            token_url=self.token_url,
            timeout=5.0,
            verify_ssl=True,
        )
        self.domino = AuthenticatedClient(
            base_url=f"{domino_host}/v4",
            api_key=self.api_key,
            token_file=self.token_file,
            token_url=self.token_url,
            headers=ACCEPT_HEADERS,
            timeout=20.0,
            verify_ssl=True,
        )

    def _set_proxy(self):
        client_source = os.getenv(DOMINO_CLIENT_SOURCE, "Python")
        flight_host = os.getenv(DOMINO_DATASOURCE_PROXY_FLIGHT_HOST)
        run_id = os.getenv(DOMINO_RUN_ID, "")
        self.proxy = flight.FlightClient(
            flight_host,
            middleware=[
                AuthMiddlewareFactory(
                    self.api_key,
                    self.token_file,
                    self.token_url,
                ),
                MetaMiddlewareFactory(client_source=client_source, run_id=run_id),
            ],
        )

    def get_datasource(self, name: str) -> Datasource:
        """Fetch a datasource by name.

        Args:
            name: unique identifier of a datasource

        Returns:
            Datasource entity with given name

        Raises:
            Exception: If the response from Domino is not 200
        """
        logger.info("get_datasource", datasource_name=name)

        run_id = os.getenv(DOMINO_RUN_ID)
        response = get_datasource_by_name.sync_detailed(
            name,
            client=self.domino,
            run_id=run_id,
        )
        if response.status_code == 200:
            datasource_dto = cast(DatasourceDto, response.parsed)
            datasource_klass = cast(
                Datasource, find_datasource_klass(datasource_dto.data_source_type)
            )
            return datasource_klass.from_dto(self, datasource_dto)

        error_response = cast(ErrorResponse, response.parsed)
        message = (
            f"Received unexpected response while getting data source: {response}"
            if error_response is None
            else error_response.message
        )
        logger.exception(message)
        raise Exception(message)

    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
    def list_keys(
        self,
        datasource_id: str,
        prefix: str,
        page_size: int,
        config: Dict[str, str],
        credential: Dict[str, str],
    ) -> List[str]:
        """List keys in a datasource.

        Args:
            datasource_id: unique identifier of a datasource
            prefix: prefix to filter keys with
            page_size: number of objects to return
            config: overwrite configuration dictionary
            credential: overwrite credential dictionary

        Returns:
            List of keys as string

        Raises:
            Exception: if the response from the Proxy is not 200
            UnauthenticatedError: if the request has invalid authentication
        """
        logger.info("list_keys", datasource_id=datasource_id, prefix=prefix, page_size=page_size)

        response = list_keys.sync_detailed(
            client=self.proxy_http,
            json_body=ListRequest(
                datasource_id=datasource_id,
                prefix=prefix,
                page_size=page_size,
                config_overwrites=APIConfig.from_dict(config),
                credential_overwrites=APIConfig.from_dict(credential),
            ),
        )

        if response.status_code == 200:
            return cast(List[str], response.parsed)
        if response.status_code == 401:
            raise UnauthenticatedError

        error = cast(ProxyErrorResponse, response.parsed)
        logger.exception(error)
        raise Exception(f"{error.error_type}: {error.raw_error}")

    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
    def get_key_url(  # pylint: disable=too-many-arguments
        self,
        datasource_id: str,
        object_key: str,
        is_read_write: bool,
        config: Dict[str, str],
        credential: Dict[str, str],
    ) -> str:
        """Request a signed URL for a given datasource and object key.

        Args:
            datasource_id: unique identifier of a datasource
            object_key: unique identifier of key to retrieve
            is_read_write: whether the signed URL allows write or not.
            config: overwrite configuration dictionary
            credential: overwrite credential dictionary

        Returns:
            Signed URL of the requested object.

        Raises:
            Exception: if the response from the Proxy is not 200
            UnauthenticatedError: if the request has invalid authentication
        """
        logger.info("get_key_url", datasource_id=datasource_id, object_key=object_key)

        response = get_key_url.sync_detailed(
            client=self.proxy_http,
            json_body=KeyRequest(
                datasource_id=datasource_id,
                object_key=object_key,
                is_read_write=is_read_write,
                config_overwrites=APIConfig.from_dict(config),
                credential_overwrites=APIConfig.from_dict(credential),
            ),
        )

        if response.status_code == 200:
            return cast(str, response.parsed)
        if response.status_code == 401:
            raise UnauthenticatedError

        error = cast(ProxyErrorResponse, response.parsed)
        logger.exception(error)
        raise Exception(f"{error.error_type}: {error.raw_error}")

    def _log_metric(
        self,
        datasource_type: str,
        bytes_processed: int,
        is_read_write: bool,
    ) -> None:
        """Log metric about file size being processed.

        Args:
            datasource_type: type of datasource
            bytes_processed: count of read or written bytes
            is_read_write: whether used for read or write
        """
        mode = LogMetricM.WRITE if is_read_write else LogMetricM.READ
        type_map = {
            DatasourceDtoDataSourceType.ADLSCONFIG.value: LogMetricT.ADLSCONFIG,
            DatasourceDtoDataSourceType.GCSCONFIG.value: LogMetricT.GCSCONFIG,
            DatasourceDtoDataSourceType.GENERICS3CONFIG.value: LogMetricT.GENERICS3CONFIG,  # noqa
            DatasourceDtoDataSourceType.S3CONFIG.value: LogMetricT.S3CONFIG,
        }
        type_ = type_map.get(datasource_type)
        if not type_:
            return

        try:
            log_metric.sync_detailed(
                client=self.proxy_http,
                t=type_,
                b=bytes_processed,
                m=mode,
            )
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception(exc)

    def execute(
        self,
        datasource_id: str,
        query: str,
        config: Dict[str, str],
        credential: Dict[str, str],
    ) -> Result:
        """Execute a given query against a datasource.

        Args:
            datasource_id: unique identifier of a datasource
            query: SQL query to execute
            config: overwrite configuration dictionary
            credential: overwrite credential dictionary

        Returns:
            Result entity encapsulating execution response

        Raises:
            DominoError: if the proxy fails to query or return data
        """
        logger.info("execute", datasource_id=datasource_id, query=query)

        try:
            self._set_proxy()
            reader = self._do_get(
                BoardingPass(
                    datasource_id=datasource_id,
                    query=query,
                    config=config,
                    credential=credential,
                ).to_json()
            )
        except (flight.FlightError, ArrowException) as exc:
            logger.exception(exc)
            raise DominoError(_unpack_flight_error(str(exc))) from None
        return Result(self, reader, query)

    @backoff.on_exception(backoff.expo, flight.FlightUnauthenticatedError, max_time=60)
    def _do_get(self, ticket: str) -> flight.FlightStreamReader:
        return self.proxy.do_get(flight.Ticket(ticket))



"""Feature Store module."""
from typing import Any, List, Optional, cast

import json
import os

from attrs import define, field

from domino_data.auth import AuthenticatedClient
from feature_store_api_client.api.default import post_featureview, post_lock, post_unlock
from feature_store_api_client.client import Client
from feature_store_api_client.models import (
    FeatureViewRequest,
    LockFeatureStoreRequest,
    UnlockFeatureStoreRequest,
    UpsertFeatureViewsRequest,
)
from feature_store_api_client.types import Response

from .exceptions import ServerException


@define
class FeatureStoreClient:
    """API client and bindings."""

    client: Client = field(init=False, repr=False)

    api_key: Optional[str] = field(factory=lambda: os.getenv("DOMINO_USER_API_KEY"))
    token_file: Optional[str] = field(factory=lambda: os.getenv("DOMINO_TOKEN_FILE"))

    def __attrs_post_init__(self):
        domino_host = os.getenv("DOMINO_API_HOST", os.getenv("DOMINO_USER_HOST", ""))
        token_url = os.getenv("DOMINO_API_PROXY")

        self.client = cast(
            Client,
            AuthenticatedClient(
                base_url=f"{domino_host}/featurestore",
                api_key=self.api_key,
                token_file=self.token_file,
                token_url=token_url,
                headers={"Accept": "application/json"},
            ),
        )

    def post_feature_views(
        self,
        feature_views: List[FeatureViewRequest],
        commit_id: str,
        project_id: str,
    ) -> None:
        """Insert or update feature views.

        Args:
            feature_views: an array of feature views to be inserted or updated.
            commit_id: the commit hash the feature store is to be synced with
            project_id: the id of the project from where the sync operation is performed

        Raises:
            ServerException: if update fails
        """
        request = UpsertFeatureViewsRequest(
            feature_views=feature_views,
            git_commit_hash=commit_id,
            project_id=project_id,
        )
        response = post_featureview.sync_detailed(
            client=self.client,
            json_body=request,
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not create Feature Views")

    def lock(self, lock_request: LockFeatureStoreRequest) -> bool:
        """Lock the feature store

        Args:
            lock_request: the lock request

        Returns:
            True if success

        Raises:
            ServerException: if lock fails
        """
        response = post_lock.sync_detailed(
            client=self.client,
            json_body=lock_request,
        )
        if response.status_code != 200:
            _raise_response_exn(response, "could not lock feature store")
        return False if response.parsed is None else response.parsed

    def unlock(self, unlock_request: UnlockFeatureStoreRequest) -> bool:
        """UnLock the feature store

        Args:
            unlock_request: the unlock request

        Returns:
            True if success

        Raises:
            ServerException: if unlock fails
        """
        response = post_unlock.sync_detailed(
            client=self.client,
            json_body=unlock_request,
        )
        if response.status_code != 200:
            _raise_response_exn(response, "could not unlock feature store")
        return False if response.parsed is None else response.parsed


def _raise_response_exn(response: Response[Any], msg: str) -> None:
    try:
        response_json = json.loads(response.content.decode("utf8"))
        server_msg = response_json.get("errors")
    except Exception:
        server_msg = None

    raise ServerException(msg, server_msg)



"""Code generated by gen.py; DO NOT EDIT.
This file was generated by robots at
2024-01-18 15:51:53.230967"""
from typing import Any, Dict, Optional, Union

from enum import Enum
from importlib import import_module

import attr

from datasource_api_client.models import DatasourceDtoDataSourceType

CREDENTIAL_TYPE = "credential"
CONFIGURATION_TYPE = "configuration"

ELEMENT_TYPE_METADATA = "__element_type_metadata"
ELEMENT_VALUE_METADATA = "__element_value_metadata"


def _filter_config(att: Any, _: Any) -> Any:
    """Filter configuration type attributes."""
    return att.metadata.get(ELEMENT_TYPE_METADATA, "") == CONFIGURATION_TYPE


def _filter_cred(att: Any, _: Any) -> Any:
    """Filter credential type attributes."""
    return att.metadata.get(ELEMENT_TYPE_METADATA, "") == CREDENTIAL_TYPE


@attr.s
class Config:
    """Base datasource configuration."""

    def config(self) -> Dict[str, str]:
        """Get configuration as dict."""
        fields = attr.fields_dict(self.__class__)
        attrs = attr.asdict(self, filter=_filter_config)

        res = {}
        for name, val in attrs.items():
            field = fields[name]
            if val is not None:
                res[field.metadata[ELEMENT_VALUE_METADATA].value] = val
        return res

    def credential(self) -> Dict[str, str]:
        """Get credentials as dict."""
        fields = attr.fields_dict(self.__class__)
        attrs = attr.asdict(self, filter=_filter_cred)

        res = {}
        for name, val in attrs.items():
            field = fields[name]
            if val is not None:
                res[field.metadata[ELEMENT_VALUE_METADATA].value] = val
        return res


class ConfigElem(Enum):
    """Enumeration of valid config elements."""

    ACCOUNTNAME = "accountName"
    ACCOUNTID = "accountID"
    BUCKET = "bucket"
    CATALOG = "catalog"
    CLUSTER = "cluster"
    DATABASE = "database"
    EXTRAPROPERTIES = "extraProperties"
    HOST = "host"
    HTTPPATH = "httpPath"
    NETWORKINGPROXY = "networkingProxy"
    PORT = "port"
    PROJECT = "project"
    REGION = "region"
    ROLE = "role"
    SCHEMA = "schema"
    SUBFOLDER = "subfolder"
    WAREHOUSE = "warehouse"
    CATALOGCODE = "catalogCode"
    ENVIRONMENT = "environment"


class CredElem(Enum):
    """Enumeration of valid credential elements."""

    ACCESSKEY = "accessKey"
    ACCESSKEYID = "accessKeyID"
    PASSWORD = "password"
    PRIVATEKEY = "privateKey"
    SECRETACCESSKEY = "secretAccessKey"
    SESSIONTOKEN = "sessionToken"
    TOKEN = "token"
    USERNAME = "username"
    CLIENTID = "clientId"
    CLIENTSECRET = "clientSecret"
    PERSONALACCESSTOKEN = "personalAccessToken"
    APIKEY = "apiKey"
    OAUTHTOKEN = "oAuthToken"


def _cred(elem: CredElem) -> Any:
    """Type helper for credentials attributes."""
    metadata = {
        ELEMENT_TYPE_METADATA: CREDENTIAL_TYPE,
        ELEMENT_VALUE_METADATA: elem,
    }
    return attr.ib(default=None, kw_only=True, metadata=metadata)


def _config(elem: ConfigElem) -> Any:
    """Type helper for configuration attributes."""
    metadata = {
        ELEMENT_TYPE_METADATA: CONFIGURATION_TYPE,
        ELEMENT_VALUE_METADATA: elem,
    }
    return attr.ib(default=None, kw_only=True, metadata=metadata)


@attr.s(auto_attribs=True)
class ADLSConfig(Config):
    """ADLSConfig datasource configuration."""

    container: Optional[str] = _config(elem=ConfigElem.BUCKET)

    access_key: Optional[str] = _cred(elem=CredElem.ACCESSKEY)


@attr.s(auto_attribs=True)
class AzureBlobStorageConfig(Config):
    """AzureBlobStorageConfig datasource configuration."""

    container: Optional[str] = _config(elem=ConfigElem.BUCKET)

    access_key: Optional[str] = _cred(elem=CredElem.ACCESSKEY)


@attr.s(auto_attribs=True)
class BigQueryConfig(Config):
    """BigQueryConfig datasource configuration."""

    gcp_project_id: Optional[str] = _config(elem=ConfigElem.PROJECT)

    private_key_json: Optional[str] = _cred(elem=CredElem.PRIVATEKEY)


@attr.s(auto_attribs=True)
class ClickHouseConfig(Config):
    """ClickHouseConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class DatabricksConfig(Config):
    """DatabricksConfig datasource configuration."""

    catalog: Optional[str] = _config(elem=ConfigElem.CATALOG)
    schema: Optional[str] = _config(elem=ConfigElem.SCHEMA)

    personal_access_token: Optional[str] = _cred(elem=CredElem.PERSONALACCESSTOKEN)


@attr.s(auto_attribs=True)
class DB2Config(Config):
    """DB2Config datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class DruidConfig(Config):
    """DruidConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class GCSConfig(Config):
    """GCSConfig datasource configuration."""

    bucket: Optional[str] = _config(elem=ConfigElem.BUCKET)

    private_key_json: Optional[str] = _cred(elem=CredElem.PRIVATEKEY)


@attr.s(auto_attribs=True)
class GenericJDBCConfig(Config):
    """GenericJDBCConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class GenericS3Config(Config):
    """GenericS3Config datasource configuration."""

    bucket: Optional[str] = _config(elem=ConfigElem.BUCKET)
    subfolder: Optional[str] = _config(elem=ConfigElem.SUBFOLDER)
    host: Optional[str] = _config(elem=ConfigElem.HOST)
    region: Optional[str] = _config(elem=ConfigElem.REGION)

    aws_access_key_id: Optional[str] = _cred(elem=CredElem.ACCESSKEYID)
    aws_secret_access_key: Optional[str] = _cred(elem=CredElem.SECRETACCESSKEY)


@attr.s(auto_attribs=True)
class GreenplumConfig(Config):
    """GreenplumConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class IgniteConfig(Config):
    """IgniteConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class MariaDBConfig(Config):
    """MariaDBConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class MongoDBConfig(Config):
    """MongoDBConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class MySQLConfig(Config):
    """MySQLConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
    region: Optional[str] = _config(elem=ConfigElem.REGION)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
    aws_access_key_id: Optional[str] = _cred(elem=CredElem.ACCESSKEYID)
    aws_secret_access_key: Optional[str] = _cred(elem=CredElem.SECRETACCESSKEY)
    aws_session_token: Optional[str] = _cred(elem=CredElem.SESSIONTOKEN)


@attr.s(auto_attribs=True)
class NetezzaConfig(Config):
    """NetezzaConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class OracleConfig(Config):
    """OracleConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class PalantirConfig(Config):
    """PalantirConfig datasource configuration."""

    client_id: Optional[str] = _cred(elem=CredElem.CLIENTID)
    client_secret: Optional[str] = _cred(elem=CredElem.CLIENTSECRET)
    o_auth_token: Optional[str] = _cred(elem=CredElem.OAUTHTOKEN)


@attr.s(auto_attribs=True)
class PostgreSQLConfig(Config):
    """PostgreSQLConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
    region: Optional[str] = _config(elem=ConfigElem.REGION)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
    aws_access_key_id: Optional[str] = _cred(elem=CredElem.ACCESSKEYID)
    aws_secret_access_key: Optional[str] = _cred(elem=CredElem.SECRETACCESSKEY)
    aws_session_token: Optional[str] = _cred(elem=CredElem.SESSIONTOKEN)


@attr.s(auto_attribs=True)
class RedshiftConfig(Config):
    """RedshiftConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
    region: Optional[str] = _config(elem=ConfigElem.REGION)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
    aws_access_key_id: Optional[str] = _cred(elem=CredElem.ACCESSKEYID)
    aws_secret_access_key: Optional[str] = _cred(elem=CredElem.SECRETACCESSKEY)
    aws_session_token: Optional[str] = _cred(elem=CredElem.SESSIONTOKEN)


@attr.s(auto_attribs=True)
class S3Config(Config):
    """S3Config datasource configuration."""

    bucket: Optional[str] = _config(elem=ConfigElem.BUCKET)
    subfolder: Optional[str] = _config(elem=ConfigElem.SUBFOLDER)
    region: Optional[str] = _config(elem=ConfigElem.REGION)

    aws_access_key_id: Optional[str] = _cred(elem=CredElem.ACCESSKEYID)
    aws_secret_access_key: Optional[str] = _cred(elem=CredElem.SECRETACCESSKEY)
    aws_session_token: Optional[str] = _cred(elem=CredElem.SESSIONTOKEN)


@attr.s(auto_attribs=True)
class SAPHanaConfig(Config):
    """SAPHanaConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class SingleStoreConfig(Config):
    """SingleStoreConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class SQLServerConfig(Config):
    """SQLServerConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class SnowflakeConfig(Config):
    """SnowflakeConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
    role: Optional[str] = _config(elem=ConfigElem.ROLE)
    schema: Optional[str] = _config(elem=ConfigElem.SCHEMA)
    warehouse: Optional[str] = _config(elem=ConfigElem.WAREHOUSE)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)
    token: Optional[str] = _cred(elem=CredElem.TOKEN)


@attr.s(auto_attribs=True)
class SynapseConfig(Config):
    """SynapseConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class TabularS3GlueConfig(Config):
    """TabularS3GlueConfig datasource configuration."""

    database: Optional[str] = _config(elem=ConfigElem.DATABASE)
    region: Optional[str] = _config(elem=ConfigElem.REGION)


@attr.s(auto_attribs=True)
class TeradataConfig(Config):
    """TeradataConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class TrinoConfig(Config):
    """TrinoConfig datasource configuration."""

    catalog: Optional[str] = _config(elem=ConfigElem.CATALOG)
    schema: Optional[str] = _config(elem=ConfigElem.SCHEMA)

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


@attr.s(auto_attribs=True)
class VerticaConfig(Config):
    """VerticaConfig datasource configuration."""

    username: Optional[str] = _cred(elem=CredElem.USERNAME)
    password: Optional[str] = _cred(elem=CredElem.PASSWORD)


DatasourceConfig = Union[
    ADLSConfig,
    AzureBlobStorageConfig,
    BigQueryConfig,
    ClickHouseConfig,
    DatabricksConfig,
    DB2Config,
    DruidConfig,
    GCSConfig,
    GenericJDBCConfig,
    GenericS3Config,
    GreenplumConfig,
    IgniteConfig,
    MariaDBConfig,
    MongoDBConfig,
    MySQLConfig,
    NetezzaConfig,
    OracleConfig,
    PalantirConfig,
    PostgreSQLConfig,
    RedshiftConfig,
    S3Config,
    SAPHanaConfig,
    SingleStoreConfig,
    SQLServerConfig,
    SnowflakeConfig,
    SynapseConfig,
    TabularS3GlueConfig,
    TeradataConfig,
    TrinoConfig,
    VerticaConfig,
    Config,
]

DATASOURCES = {
    DatasourceDtoDataSourceType.ADLSCONFIG: "ObjectStoreDatasource",
    DatasourceDtoDataSourceType.AZUREBLOBSTORAGECONFIG: "ObjectStoreDatasource",
    DatasourceDtoDataSourceType.BIGQUERYCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.CLICKHOUSECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.DATABRICKSCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.DB2CONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.DRUIDCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.GCSCONFIG: "ObjectStoreDatasource",
    DatasourceDtoDataSourceType.GENERICJDBCCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.GENERICS3CONFIG: "ObjectStoreDatasource",
    DatasourceDtoDataSourceType.GREENPLUMCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.IGNITECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.MARIADBCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.MONGODBCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.MYSQLCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.NETEZZACONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.ORACLECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.PALANTIRCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.POSTGRESQLCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.REDSHIFTCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.S3CONFIG: "ObjectStoreDatasource",
    DatasourceDtoDataSourceType.SAPHANACONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.SINGLESTORECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.SQLSERVERCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.SNOWFLAKECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.SYNAPSECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.TABULARS3GLUECONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.TERADATACONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.TRINOCONFIG: "TabularDatasource",
    DatasourceDtoDataSourceType.VERTICACONFIG: "TabularDatasource",
}


def find_datasource_klass(datasource_type: DatasourceDtoDataSourceType) -> Any:
    """Find according datasource class."""
    mod = import_module("domino_data.data_sources")
    return getattr(mod, DATASOURCES[datasource_type])



class DominoError(Exception):
    """Base exception for known errors."""


class ServerException(Exception):
    """This exception is raised when the FeatureStore server rejects a request."""

    def __init__(self, message: str, server_msg: str):
        self.message = message
        self.server_msg = server_msg


class FeastRepoError(Exception):
    """Raised when no feast repo is found or more than one feast repos are found"""


class GitPullError(Exception):
    """Raised when git pull failed"""


class GitPushError(Exception):
    """Raised when git push failed"""


class FeatureStoreLockError(Exception):
    """Raised when failed to lock or unlock the feature store"""



from enum import Enum


class DatasourceDtoAuthType(str, Enum):
    AZUREBASIC = "AzureBasic"
    BASIC = "Basic"
    AWSIAMBASIC = "AWSIAMBasic"
    AWSIAMBASICNOOVERRIDE = "AWSIAMBasicNoOverride"
    AWSIAMROLE = "AWSIAMRole"
    AWSIAMROLEWITHUSERNAME = "AWSIAMRoleWithUsername"
    GCPBASIC = "GCPBasic"
    OAUTH = "OAuth"
    OAUTHTOKEN = "OAuthToken"
    CLIENTIDSECRET = "ClientIdSecret"
    PERSONALTOKEN = "PersonalToken"

    def __str__(self) -> str:
        return str(self.value)



""" Classes to augment headers with metadata for HTTP and Flight requests."""

from typing import Optional

import attr
from pyarrow import flight


@attr.s(auto_attribs=True)
class MetaMiddlewareFactory(flight.ClientMiddlewareFactory):
    """Middleware Factory for metadata."""

    client_source: Optional[str] = attr.ib()
    run_id: Optional[str] = attr.ib()

    def start_call(self, _):
        """Called at the start of an RPC."""
        return MetaMiddleware(self.client_source, self.run_id)


@attr.s(auto_attribs=True)
class MetaMiddleware(flight.ClientMiddleware):
    """Middleware for authenticating flight requests."""

    client_source: Optional[str] = attr.ib()
    run_id: Optional[str] = attr.ib()

    def call_completed(self, _):
        """No implementation. TODO logging."""

    def received_headers(self, _):
        """No implementation."""

    def sending_headers(self):
        """Return metadata headers."""
        headers = {}

        if self.client_source is not None:
            headers["x-domino-client-source"] = self.client_source
        if self.run_id is not None:
            headers["x-domino-run-id"] = self.run_id

        return headers



from typing import BinaryIO, Generator, Optional, Tuple

import io
import shutil
import threading
from concurrent.futures import ThreadPoolExecutor

import urllib3

MAX_WORKERS = 10
MB = 2**20  # 2^20 bytes - 1 Megabyte


def split_range(start: int, end: int, step: int) -> Generator[Tuple[int, int], None, None]:
    """Yield all steps from start to end as tuple of integers.

    All returned block are inclusive to ensure every integer is covered from start to end.

    Example:
        start: 0, end: 10, step: 3 -> (0, 2), (3, 5), (6, 8), (9, 10)
        start: 0, end: 10, step: 4 -> (0, 3), (4, 8), (9, 10)

    Args:
        start: start of the range to split
        end: end of the range to split
        step: step size

    Yields:
        tuple[int, int]
    """
    r = range(start, end, step)
    max_block = start

    for min_block, max_block in zip(r[:-1], r[1:]):
        yield (min_block, max_block - 1)

    yield (max_block, end)


class BlobTransfer:
    def __init__(
        self,
        url: str,
        destination: BinaryIO,
        max_workers: int = MAX_WORKERS,
        # Recommended chunk size by Amazon S3
        # See https://docs.aws.amazon.com/whitepapers/latest/s3-optimizing-performance-best-practices/use-byte-range-fetches.html  # noqa
        chunk_size: int = 16 * MB,
        http: Optional[urllib3.PoolManager] = None,
    ):
        self.url = url
        self.http = http or urllib3.PoolManager()
        self.destination = destination
        self.content_size = self._get_content_size()

        self._lock = threading.Lock()

        with ThreadPoolExecutor(max_workers) as pool:
            pool.map(self._get_part, split_range(0, self.content_size, chunk_size))

    def _get_content_size(self) -> int:
        res = self.http.request("GET", self.url, headers={"Range": "bytes=0-0"})
        return int(res.headers["Content-Range"].partition("/")[-1])

    def _get_part(self, block: Tuple[int, int]) -> None:
        """Download specific block of data from blob and writes it into destination.

        Args:
            block: block of bytes to download
        """
        headers = {"Range": f"bytes={block[0]}-{block[1]}"}
        res = self.http.request("GET", self.url, headers=headers, preload_content=False)

        buffer = io.BytesIO()
        shutil.copyfileobj(res, buffer)

        buffer.seek(0)
        with self._lock:
            self.destination.seek(block[0])
            shutil.copyfileobj(buffer, self.destination)  # type: ignore

        buffer.close()
        res.release_connection()



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.create_feature_store_request_offline_store_config import (
    CreateFeatureStoreRequestOfflineStoreConfig,
)
from ..models.git_provider_name import GitProviderName
from ..models.offline_store_type import OfflineStoreType

T = TypeVar("T", bound="CreateFeatureStoreRequest")


@attr.s(auto_attribs=True)
class CreateFeatureStoreRequest:
    """
    Attributes:
        name (str):
        offline_store_type (OfflineStoreType):
        offline_store_config (CreateFeatureStoreRequestOfflineStoreConfig):
        git_repo (str):
        git_service_provider (GitProviderName):
    """

    name: str
    offline_store_type: OfflineStoreType
    offline_store_config: CreateFeatureStoreRequestOfflineStoreConfig
    git_repo: str
    git_service_provider: GitProviderName
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        offline_store_type = self.offline_store_type.value

        offline_store_config = self.offline_store_config.to_dict()

        git_repo = self.git_repo
        git_service_provider = self.git_service_provider.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "offlineStoreType": offline_store_type,
                "offlineStoreConfig": offline_store_config,
                "gitRepo": git_repo,
                "gitServiceProvider": git_service_provider,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        offline_store_type = OfflineStoreType(d.pop("offlineStoreType"))

        offline_store_config = CreateFeatureStoreRequestOfflineStoreConfig.from_dict(
            d.pop("offlineStoreConfig")
        )

        git_repo = d.pop("gitRepo")

        git_service_provider = GitProviderName(d.pop("gitServiceProvider"))

        create_feature_store_request = cls(
            name=name,
            offline_store_type=offline_store_type,
            offline_store_config=offline_store_config,
            git_repo=git_repo,
            git_service_provider=git_service_provider,
        )

        create_feature_store_request.additional_properties = d
        return create_feature_store_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.training_set_filter_meta import TrainingSetFilterMeta
from ..types import UNSET, Unset

T = TypeVar("T", bound="TrainingSetFilter")


@attr.s(auto_attribs=True)
class TrainingSetFilter:
    """ """

    project_id: str
    meta: Union[Unset, TrainingSetFilterMeta] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        project_id = self.project_id
        meta: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.meta, Unset):
            meta = self.meta.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "projectId": project_id,
            }
        )
        if meta is not UNSET:
            field_dict["meta"] = meta

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        project_id = d.pop("projectId")

        _meta = d.pop("meta", UNSET)
        meta: Union[Unset, TrainingSetFilterMeta]
        if isinstance(_meta, Unset):
            meta = UNSET
        else:
            meta = TrainingSetFilterMeta.from_dict(_meta)

        training_set_filter = cls(
            project_id=project_id,
            meta=meta,
        )

        training_set_filter.additional_properties = d
        return training_set_filter

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os

_import_error_message = (
    "pinecone-client is not installed.\n\n"
    "Please pip install pinecone-client:\n\n"
    "  python -m pip install pinecone-client   # install directly\n"
)

try:
    from pinecone.core.client.configuration import Configuration as OpenApiConfiguration
except ImportError as e:
    if e.msg == "No module named 'pinecone'":
        raise ImportError(_import_error_message) from e
    else:
        raise


class DominoPineconeConfiguration(OpenApiConfiguration):
    def __init__(
        self,
        datasource=None,
        host=None,
        api_key=None,
        api_key_prefix=None,
        access_token=None,
        username=None,
        password=None,
        discard_unknown_keys=False,
        disabled_client_side_validations="",
        server_index=None,
        server_variables=None,
        server_operation_index=None,
        server_operation_variables=None,
        ssl_ca_cert=None,
    ):

        super().__init__(
            host,
            api_key,
            api_key_prefix,
            access_token,
            username,
            password,
            discard_unknown_keys,
            disabled_client_side_validations,
            server_index,
            server_variables,
            server_operation_index,
            server_operation_variables,
            ssl_ca_cert,
        )

        self.proxy = os.getenv("DOMINO_DATA_API_GATEWAY", "http://127.0.0.1:8766")
        self.proxy_headers = {"X-Domino-Datasource": datasource}

    def get_host_from_settings(self, index, variables=None, servers=None):
        url = super().get_host_from_settings(index, variables, servers)
        return url.replace("https://", "http://")



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="Field")


@attr.s(auto_attribs=True)
class Field:
    """
    Attributes:
        is_optional (bool):
        is_overridable (bool):
        name (str):
        alias (Union[Unset, str]):
        is_secret (Union[Unset, bool]):
        regexp (Union[Unset, str]):
        regexp_error_message (Union[Unset, str]):
    """

    is_optional: bool
    is_overridable: bool
    name: str
    alias: Union[Unset, str] = UNSET
    is_secret: Union[Unset, bool] = UNSET
    regexp: Union[Unset, str] = UNSET
    regexp_error_message: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        is_optional = self.is_optional
        is_overridable = self.is_overridable
        name = self.name
        alias = self.alias
        is_secret = self.is_secret
        regexp = self.regexp
        regexp_error_message = self.regexp_error_message

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "isOptional": is_optional,
                "isOverridable": is_overridable,
                "name": name,
            }
        )
        if alias is not UNSET:
            field_dict["alias"] = alias
        if is_secret is not UNSET:
            field_dict["isSecret"] = is_secret
        if regexp is not UNSET:
            field_dict["regexp"] = regexp
        if regexp_error_message is not UNSET:
            field_dict["regexpErrorMessage"] = regexp_error_message

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        is_optional = d.pop("isOptional")

        is_overridable = d.pop("isOverridable")

        name = d.pop("name")

        alias = d.pop("alias", UNSET)

        is_secret = d.pop("isSecret", UNSET)

        regexp = d.pop("regexp", UNSET)

        regexp_error_message = d.pop("regexpErrorMessage", UNSET)

        field = cls(
            is_optional=is_optional,
            is_overridable=is_overridable,
            name=name,
            alias=alias,
            is_secret=is_secret,
            regexp=regexp,
            regexp_error_message=regexp_error_message,
        )

        field.additional_properties = d
        return field

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

import attr

from ..models.entity import Entity
from ..models.feature import Feature
from ..models.feature_view_tags import FeatureViewTags
from ..models.metadata import Metadata
from ..types import UNSET, Unset

T = TypeVar("T", bound="FeatureView")


@attr.s(auto_attribs=True)
class FeatureView:
    """
    Attributes:
        id (str):
        name (str):
        feature_store_id (str):
        metadata (Metadata):
        entities (List[Entity]):
        features (List[Feature]):
        tags (FeatureViewTags):
        project_ids (List[str]):
        ttl (Union[Unset, int]):
    """

    id: str
    name: str
    feature_store_id: str
    metadata: Metadata
    entities: List[Entity]
    features: List[Feature]
    tags: FeatureViewTags
    project_ids: List[str]
    ttl: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        name = self.name
        feature_store_id = self.feature_store_id
        metadata = self.metadata.to_dict()

        entities = []
        for entities_item_data in self.entities:
            entities_item = entities_item_data.to_dict()

            entities.append(entities_item)

        features = []
        for features_item_data in self.features:
            features_item = features_item_data.to_dict()

            features.append(features_item)

        tags = self.tags.to_dict()

        project_ids = self.project_ids

        ttl = self.ttl

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
                "featureStoreId": feature_store_id,
                "metadata": metadata,
                "entities": entities,
                "features": features,
                "tags": tags,
                "projectIds": project_ids,
            }
        )
        if ttl is not UNSET:
            field_dict["ttl"] = ttl

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        name = d.pop("name")

        feature_store_id = d.pop("featureStoreId")

        metadata = Metadata.from_dict(d.pop("metadata"))

        entities = []
        _entities = d.pop("entities")
        for entities_item_data in _entities:
            entities_item = Entity.from_dict(entities_item_data)

            entities.append(entities_item)

        features = []
        _features = d.pop("features")
        for features_item_data in _features:
            features_item = Feature.from_dict(features_item_data)

            features.append(features_item)

        tags = FeatureViewTags.from_dict(d.pop("tags"))

        project_ids = cast(List[str], d.pop("projectIds"))

        ttl = d.pop("ttl", UNSET)

        feature_view = cls(
            id=id,
            name=name,
            feature_store_id=feature_store_id,
            metadata=metadata,
            entities=entities,
            features=features,
            tags=tags,
            project_ids=project_ids,
            ttl=ttl,
        )

        feature_view.additional_properties = d
        return feature_view

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, cast

import datetime

import attr
from dateutil.parser import isoparse

from ..models.feature_store_offline_store_config import FeatureStoreOfflineStoreConfig
from ..models.git_provider_name import GitProviderName
from ..models.offline_store_type import OfflineStoreType

T = TypeVar("T", bound="FeatureStore")


@attr.s(auto_attribs=True)
class FeatureStore:
    """
    Attributes:
        id (str):
        owner_id (str):
        creation_time (datetime.datetime):
        offline_store_type (OfflineStoreType):
        offline_store_config (FeatureStoreOfflineStoreConfig):
        git_repo (str):
        git_service_provider (GitProviderName):
        project_ids (List[str]):
        last_updated_time (datetime.datetime):
    """

    id: str
    owner_id: str
    creation_time: datetime.datetime
    offline_store_type: OfflineStoreType
    offline_store_config: FeatureStoreOfflineStoreConfig
    git_repo: str
    git_service_provider: GitProviderName
    project_ids: List[str]
    last_updated_time: datetime.datetime
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        owner_id = self.owner_id
        creation_time = self.creation_time.isoformat()

        offline_store_type = self.offline_store_type.value

        offline_store_config = self.offline_store_config.to_dict()

        git_repo = self.git_repo
        git_service_provider = self.git_service_provider.value

        project_ids = self.project_ids

        last_updated_time = self.last_updated_time.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "ownerId": owner_id,
                "creationTime": creation_time,
                "offlineStoreType": offline_store_type,
                "offlineStoreConfig": offline_store_config,
                "gitRepo": git_repo,
                "gitServiceProvider": git_service_provider,
                "projectIds": project_ids,
                "lastUpdatedTime": last_updated_time,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        owner_id = d.pop("ownerId")

        creation_time = isoparse(d.pop("creationTime"))

        offline_store_type = OfflineStoreType(d.pop("offlineStoreType"))

        offline_store_config = FeatureStoreOfflineStoreConfig.from_dict(d.pop("offlineStoreConfig"))

        git_repo = d.pop("gitRepo")

        git_service_provider = GitProviderName(d.pop("gitServiceProvider"))

        project_ids = cast(List[str], d.pop("projectIds"))

        last_updated_time = isoparse(d.pop("lastUpdatedTime"))

        feature_store = cls(
            id=id,
            owner_id=owner_id,
            creation_time=creation_time,
            offline_store_type=offline_store_type,
            offline_store_config=offline_store_config,
            git_repo=git_repo,
            git_service_provider=git_service_provider,
            project_ids=project_ids,
            last_updated_time=last_updated_time,
        )

        feature_store.additional_properties = d
        return feature_store

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class DatasourceDtoDataSourceType(str, Enum):
    ADLSCONFIG = "ADLSConfig"
    AZUREBLOBSTORAGECONFIG = "AzureBlobStorageConfig"
    BIGQUERYCONFIG = "BigQueryConfig"
    CLICKHOUSECONFIG = "ClickHouseConfig"
    DATABRICKSCONFIG = "DatabricksConfig"
    DB2CONFIG = "DB2Config"
    DRUIDCONFIG = "DruidConfig"
    GCSCONFIG = "GCSConfig"
    GENERICJDBCCONFIG = "GenericJDBCConfig"
    GENERICS3CONFIG = "GenericS3Config"
    GREENPLUMCONFIG = "GreenplumConfig"
    IGNITECONFIG = "IgniteConfig"
    MARIADBCONFIG = "MariaDBConfig"
    MONGODBCONFIG = "MongoDBConfig"
    MYSQLCONFIG = "MySQLConfig"
    NETEZZACONFIG = "NetezzaConfig"
    ORACLECONFIG = "OracleConfig"
    PALANTIRCONFIG = "PalantirConfig"
    POSTGRESQLCONFIG = "PostgreSQLConfig"
    REDSHIFTCONFIG = "RedshiftConfig"
    S3CONFIG = "S3Config"
    SAPHANACONFIG = "SAPHanaConfig"
    SINGLESTORECONFIG = "SingleStoreConfig"
    SQLSERVERCONFIG = "SQLServerConfig"
    SNOWFLAKECONFIG = "SnowflakeConfig"
    SYNAPSECONFIG = "SynapseConfig"
    TABULARS3GLUECONFIG = "TabularS3GlueConfig"
    TERADATACONFIG = "TeradataConfig"
    TRINOCONFIG = "TrinoConfig"
    VERTICACONFIG = "VerticaConfig"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar

import attr

from ..models.auth_type import AuthType
from ..models.offline_store_config_fields import OfflineStoreConfigFields
from ..models.offline_store_type import OfflineStoreType

T = TypeVar("T", bound="OfflineStoreConfig")


@attr.s(auto_attribs=True)
class OfflineStoreConfig:
    """
    Attributes:
        auth_types (List[AuthType]):
        offline_store_type (OfflineStoreType):
        fields (OfflineStoreConfigFields):
    """

    auth_types: List[AuthType]
    offline_store_type: OfflineStoreType
    fields: OfflineStoreConfigFields
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        auth_types = []
        for auth_types_item_data in self.auth_types:
            auth_types_item = auth_types_item_data.value

            auth_types.append(auth_types_item)

        offline_store_type = self.offline_store_type.value

        fields = self.fields.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "authTypes": auth_types,
                "offlineStoreType": offline_store_type,
                "fields": fields,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        auth_types = []
        _auth_types = d.pop("authTypes")
        for auth_types_item_data in _auth_types:
            auth_types_item = AuthType(auth_types_item_data)

            auth_types.append(auth_types_item)

        offline_store_type = OfflineStoreType(d.pop("offlineStoreType"))

        fields = OfflineStoreConfigFields.from_dict(d.pop("fields"))

        offline_store_config = cls(
            auth_types=auth_types,
            offline_store_type=offline_store_type,
            fields=fields,
        )

        offline_store_config.additional_properties = d
        return offline_store_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Optional, Union

import httpx

from ...client import Client
from ...models.datasource_dto import DatasourceDto
from ...models.error_response import ErrorResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    name: str,
    *,
    client: Client,
    run_id: Union[Unset, None, str] = UNSET,
) -> Dict[str, Any]:
    url = "{}/datasource/name/{name}".format(client.base_url, name=name)

    headers: Dict[str, str] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    params: Dict[str, Any] = {}
    params["runId"] = run_id

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "method": "get",
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "params": params,
    }


def _parse_response(*, response: httpx.Response) -> Optional[Union[DatasourceDto, ErrorResponse]]:
    if response.status_code == 200:
        response_200 = DatasourceDto.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = ErrorResponse.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = ErrorResponse.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = ErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == 500:
        response_500 = ErrorResponse.from_dict(response.json())

        return response_500
    return None


def _build_response(*, response: httpx.Response) -> Response[Union[DatasourceDto, ErrorResponse]]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    name: str,
    *,
    client: Client,
    run_id: Union[Unset, None, str] = UNSET,
) -> Response[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, None, str]):

    Returns:
        Response[Union[DatasourceDto, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        name=name,
        client=client,
        run_id=run_id,
    )

    response = httpx.request(
        verify=client.verify_ssl,
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    name: str,
    *,
    client: Client,
    run_id: Union[Unset, None, str] = UNSET,
) -> Optional[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, None, str]):

    Returns:
        Response[Union[DatasourceDto, ErrorResponse]]
    """

    return sync_detailed(
        name=name,
        client=client,
        run_id=run_id,
    ).parsed


async def asyncio_detailed(
    name: str,
    *,
    client: Client,
    run_id: Union[Unset, None, str] = UNSET,
) -> Response[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, None, str]):

    Returns:
        Response[Union[DatasourceDto, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        name=name,
        client=client,
        run_id=run_id,
    )

    async with httpx.AsyncClient(verify=client.verify_ssl) as _client:
        response = await _client.request(**kwargs)

    return _build_response(response=response)


async def asyncio(
    name: str,
    *,
    client: Client,
    run_id: Union[Unset, None, str] = UNSET,
) -> Optional[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, None, str]):

    Returns:
        Response[Union[DatasourceDto, ErrorResponse]]
    """

    return (
        await asyncio_detailed(
            name=name,
            client=client,
            run_id=run_id,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..types import UNSET, Unset

T = TypeVar("T", bound="ProxyErrorResponse")


@attr.s(auto_attribs=True)
class ProxyErrorResponse:
    """
    Attributes:
        error_type (Union[Unset, str]):
        sub_type (Union[Unset, str]):
        raw_error (Union[Unset, str]):
    """

    error_type: Union[Unset, str] = UNSET
    sub_type: Union[Unset, str] = UNSET
    raw_error: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        error_type = self.error_type
        sub_type = self.sub_type
        raw_error = self.raw_error

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if error_type is not UNSET:
            field_dict["errorType"] = error_type
        if sub_type is not UNSET:
            field_dict["subType"] = sub_type
        if raw_error is not UNSET:
            field_dict["rawError"] = raw_error

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        error_type = d.pop("errorType", UNSET)

        sub_type = d.pop("subType", UNSET)

        raw_error = d.pop("rawError", UNSET)

        proxy_error_response = cls(
            error_type=error_type,
            sub_type=sub_type,
            raw_error=raw_error,
        )

        proxy_error_response.additional_properties = d
        return proxy_error_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Domino Data API for interacting with Access Data features"""

from importlib import metadata as importlib_metadata

# Pyarrow CVE hotfix https://lists.apache.org/thread/4notgcj3y7j5z4vxcr6o966g52jqxpdt
import pyarrow_hotfix  # noqa: F401


def get_version() -> str:
    """Get installed packaged version."""
    try:
        return importlib_metadata.version("dominodatalab-data")
    except importlib_metadata.PackageNotFoundError:
        return "unknown"


version: str = get_version()
